#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#if ENABLE_NLS
#include <libintl.h>
#define _(String) gettext( String )
#else
#define _(String) String
#endif
#define N_(String) String
#if HAVE_SETLOCALE
#include <locale.h>
#endif

#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <netdb.h>
#include <syslog.h>
#include <signal.h>
#ifdef __vax__
#include <sys/file.h>
#include <sys/socket.h>
#else
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/param.h>
#include <net/if.h>
#include <protocols/rwhod.h>
#include "rwhod.h"

#ifndef _PATH_RWHODIR
#define _PATH_RWHODIR "/var/spool/rwho"
#endif
#ifndef _PATH_DEVNULL
#define _PATH_DEVNULL "/dev/null"
#endif

#define	HOSTSINC	16

static	int	opt_nodetach;
static	int	opt_verbose;
static	int	opt_nobroadcast;
static	int	opt_pointopoint;
static	int	opt_interval = 5*60;
static	int	show_version;
static	int	show_help;

static	char	*progname;

static	int	sock;
static	struct	sockaddr_in *hosts;
static	int	nrhosts;
static	int	maxhosts;
static	int	service;
static	struct	whod rwho;

#define	LAST(var)	(var+sizeof(var))

#define MAKEROOM(base,cnt,max,inc)	\
	if ((cnt) == (max)) {		\
		(max) += (inc);		\
		(base) = realloc((base),(max)*sizeof((base)[0]) ); \
		if ((base) == NULL) {	\
			printf(_("%s: out of memory\n"), progname);	\
			exit(1);	\
		}			\
	}

static
void usage(int status)
{
	if (status != 0) {
		printf(_("Try `%s --help' for more information.\n"),
			progname);
	}
	else {
		printf(_("Usage: %s [OPTION] [HOST]...\n"),
			progname);
		printf(_("Send host status to remote machines.\n"
			"\n"
			"  -d  --no-detach      no not detach\n"
			"  -v  --verbose        increase verbosity level (implicit -d)\n"
			"  -b  --no-broadcast   do not include broadcast interfaces\n"
			"  -t  --pointopoint    include point-to-point interfaces\n"
			"  -i  --interval       interval at which we send our statistics\n"
			"  -p  --port           port at which we send/listen for statistics\n"
			"      --help           display this help and exit\n"
			"      --version        output version information and exit\n"
			"\n"
			"Each HOST mentioned on the commandline will be send the hoststatus\n"
			"directly instead of a broadcast. The default interval is 5 minutes.\n"));
	}
	exit(status);
}

void sigsent(int sig)
{
	int			i;
	int			len;
	struct	ifconf		ifc;
	struct	ifreq		ifr[24];
	struct	sockaddr_in	sin;

	signal(sig,sigsent);
	alarm(opt_interval);

	/* update the whod structure */
	if ((len=updatewhod(&rwho)) == 0)
		return;

	/* add the broadcast devices when required */
	ifc.ifc_len = sizeof(ifr);
	ifc.ifc_req = ifr;
	if (ioctl(sock,SIOCGIFCONF,&ifc) != 0) {
		syslog(LOG_CRIT, _("%s: unable to get the network interfaces\n"),progname);
		exit(1);
	}

	/*
         * dynamicly find the allowed interfaces
	 */
	for (i=0; ifc.ifc_len>0; i++,ifc.ifc_len -= sizeof(struct ifreq)) {
		if (ioctl(sock,SIOCGIFFLAGS,&ifr[i]) != 0) {
			syslog(LOG_WARNING, _("%s: unable to get interface flags on %s\n"),
				progname, ifr[i].ifr_name);
			continue;
		}

		/* loopback or down devices are skipped altogether */
		if ((ifr[i].ifr_flags & (IFF_LOOPBACK|IFF_UP)) != IFF_UP)
			continue;

		/* is it a broadcast interface? */
		if (ifr[i].ifr_flags & IFF_BROADCAST) {
			if (opt_nobroadcast)
				continue;	/* not allowed */

			if (ioctl(sock,SIOCGIFBRDADDR,&ifr[i]) != 0) {
				syslog(LOG_WARNING, _("%s: unable to get interface broadcast address\n"),
					progname, ifr[i].ifr_name);
				continue;
			}
			memcpy(&sin,&ifr[i].ifr_broadaddr,sizeof(struct sockaddr));
		}
		/* is it a point-to-point interface? */
		else if (ifr[i].ifr_flags & IFF_POINTOPOINT) {
			if (!opt_pointopoint)
				continue;	/* not allowed */

			if (ioctl(sock,SIOCGIFDSTADDR,&ifr[i]) != 0) {
				syslog(LOG_WARNING, _("%s: unable to get interface destination address\n"),
					progname, ifr[i].ifr_name);
				continue;
			}
			memcpy(&sin,&ifr[i].ifr_dstaddr,sizeof(struct sockaddr));
		}
		/* than it must be a normal device */
		else {
			if (ioctl(sock,SIOCGIFADDR,&ifr[i]) != 0) {
				syslog(LOG_WARNING,_("%s: unable to get interface address\n"),
					progname, ifr[i].ifr_name);
				continue;
			}
			memcpy(&sin,&ifr[i].ifr_addr,sizeof(struct sockaddr));
		}

		/* these two need not be valid - better make sure */
		sin.sin_family = AF_INET;
		sin.sin_port = service;

		if (opt_verbose)
			printf(_("send(%d) to %s:%d\n"),len,inet_ntoa(sin.sin_addr),ntohs(sin.sin_port));
		if (sendto(sock,&rwho,len,0,(struct sockaddr *)&sin,sizeof(struct sockaddr)) != len)
			syslog(LOG_CRIT, _("%s: send failed (%m)\n"),progname);
	}

	/* send data to recipitants */
	for (i=0; i<nrhosts; i++) {
		if (opt_verbose)
			printf(_("send(%d) to %s:%d\n"),len,inet_ntoa(hosts[i].sin_addr),ntohs(hosts[i].sin_port));
		if (sendto(sock,&rwho,len,0,(struct sockaddr *)&hosts[i],sizeof(struct sockaddr)) != len)
			syslog(LOG_CRIT, _("%s: send failed (%m)\n"),progname);
	}
}

int main(int argc, char *argv[])
{
static struct option longopt[] = {
	{ "no-broadcast", no_argument, &opt_nobroadcast, 1},
	{ "pointopoint", no_argument, &opt_pointopoint, 1},
	{ "no-detach", no_argument, &opt_nodetach, 1},
	{ "verbose", no_argument, NULL, 'v'},
	{ "interval", required_argument, NULL, 1},
	{ "port", required_argument, NULL, 'p'},
	{ "help", no_argument, &show_help, 1},
	{ "version", no_argument, &show_version, 1},
	{ NULL } };
	struct	sockaddr_in	sin;
	struct	servent		*sp;
	int			i;
	int			on = 1;

#if HAVE_SETLOCALE
	setlocale(LC_MESSAGES, "");
#endif
#if ENABLE_NLS
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif

	progname = strrchr(argv[0],'/');
	if (progname) progname++;
	else progname = argv[0];

	sp = getservbyname("who","udp");
	if (sp) service = sp->s_port;
	else service = htons(IPPORT_WHOSERVER);
	endservent();

	while ((i=getopt_long(argc,argv,"btdvsi:p:",longopt,NULL)) != -1)
	 switch (i) {
	  case  0  :	/* long option */
	  		break;
	  case 'b' :	opt_nobroadcast = 1;
			break;
	  case 't' :	opt_pointopoint = 1;
			break;
	  case 'd' :	opt_nodetach = 1;
			break;
	  case 'v' :	opt_verbose++;
			opt_nodetach = 1;
			break;
	  case  1  :	/* fall through */
	  case 'i' :	if ((i=atoi(optarg))<0) {
				printf(_("%s: interval must be numeric\n"),progname);
				usage(1);
			}
			opt_interval = i*60;
			break;
	  case 'p' :	/* port */
			if ((i=atoi(optarg))<0) {
				printf(_("%s: port must be numeric\n"),progname);
				usage(1);
			}
			service = htons(i);
			break;
	  case '?' :	usage(1);
			break;
	 }

	/* the rest of the commandline must be hostsnames */
	for (; optind<argc; optind++) {
		struct	hostent	*hp;

		MAKEROOM(hosts,nrhosts,maxhosts,HOSTSINC);

		hosts[nrhosts].sin_family = AF_INET;
		hosts[nrhosts].sin_port = service;
#if HAVE_INET_ATON
		if (inet_aton(argv[optind],&hosts[nrhosts].sin_addr) == 0) {
#else
		if ((hosts[nrhosts].sin_addr.s_addr = inet_addr(argv[optind])) == -1) {
#endif
			hp = gethostbyname(argv[optind]);
			if (hp==NULL) {
				printf(_("%s: unknown host %s\n"),progname,argv[optind]);
				usage(1);
			}
			memcpy(&hosts[nrhosts].sin_addr.s_addr,hp->h_addr,hp->h_length);
		}
		nrhosts++;
	}
	endhostent();

	if (show_version) {
		printf("%s - %s\n", progname, PACKAGE_VERSION);
		exit(0);
	}

	if (show_help)
		usage(0);

	if (!initwhod(&rwho)) {
		printf(_("%s: unable to acquire rwhod data\n"),progname);
		exit(1);
	}

	sock = socket(AF_INET,SOCK_DGRAM,0);
	if (sock<0) {
		perror(progname);
		exit(1);
	}
	if (setsockopt(sock,SOL_SOCKET,SO_BROADCAST,&on,sizeof(on)) != 0) {
		perror(progname);
		exit(1);
	}

	/* bind() before we go to al the trouble of becoming a daemon */
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_port = service;
	if (bind(sock,(struct sockaddr *)&sin,sizeof(sin)) != 0) {
		perror(progname);
		exit(1);
	}

	/* lets pretent to be a daemon and detach */
	if (!opt_nodetach) {
		switch (fork()) {
		 int fd;
		 case -1 :
			/* an error?? */
			perror(progname);
			exit(1);
		 case  0 :
			/* child */
			setsid(); /* no controlling tty */
			umask(0); /* files must be readable by others */
			chdir("/"); /* no directory */
			setuid(geteuid());
			setgid(getegid());
			close(0); /* no files open */
			close(1);
			close(2);
			fd = open(_PATH_DEVNULL, O_RDWR);
			if (fd >= 0) {
				if (fd != 0) {
					dup2(fd,0); /* stdin < /dev/null */
					close(fd);
				}
				dup2(0,1); 	/* stdout > /dev/null */
				dup2(0,2);	/* stderr > /dev/null */
			}
			break;
		 default :
			/* dady has nothing to say... */
			exit(0);
		}
	}

	/* setup a signal handler */
	signal(SIGALRM,sigsent);
	alarm(opt_interval);

	/* open access to the syslog files */
	openlog("rwhod",LOG_PID,LOG_DAEMON);

	/* do a forced firsttime sent */
	sigsent(SIGALRM);

	/* wait for data to come in... */
	for (;;) {
		int		len, size;
		int		fd;
		struct whod	rwho;
		char		nm[MAXPATHLEN];
		char		*p;
		time_t		now;

		/* note: we protect outself from arraybound attacks
		   by giving recvfrom a maximum size to work with */
		size = sizeof(sin);
		len = size = recvfrom(sock,&rwho,sizeof(rwho),0,(struct sockaddr *)&sin, &size);
		if (len<0)
			continue;

		if (len<60) {
			syslog(LOG_WARNING,_("Message too short (%d)"),len);
			continue;
		}
		if (sin.sin_port != service) {
			syslog(LOG_WARNING,_("Data from port (%d)"),ntohs(sin.sin_port));
			continue;
		}
		p = rwho.wd_hostname;
		while (*p && p!=LAST(rwho.wd_hostname) && isprint(*p))
			p++;
		if (p == LAST(rwho.wd_hostname)) {
			syslog(LOG_WARNING,_("Unterminated hostname from %s"),inet_ntoa(sin.sin_addr));
			*--p = '\0';
		}
		if (*p) {
			syslog(LOG_WARNING,_("Non-printable hostname from %s"),inet_ntoa(sin.sin_addr));
			continue;
		}

		/* convert received data to host format */
		time(&now);
		rwho.wd_recvtime = now;
		rwho.wd_sendtime = ntohl(rwho.wd_sendtime);
		rwho.wd_boottime = ntohl(rwho.wd_boottime);
		rwho.wd_loadav[0] = ntohl(rwho.wd_loadav[0]);
		rwho.wd_loadav[1] = ntohl(rwho.wd_loadav[1]);
		rwho.wd_loadav[2] = ntohl(rwho.wd_loadav[2]);
		for (i=0; size>60; i++,size -= sizeof(struct whoent)) {
			rwho.wd_we[i].we_idle = ntohl(rwho.wd_we[i].we_idle);
			rwho.wd_we[i].we_utmp.out_time = ntohl(rwho.wd_we[i].we_utmp.out_time);
		}
		if (opt_verbose)
			printf(_("recv(%d) from %s (%s:%d)\n"),len,
				rwho.wd_hostname,
				inet_ntoa(sin.sin_addr),
				ntohs(sin.sin_port));
		sprintf(nm,"%s/whod.%s",_PATH_RWHODIR,rwho.wd_hostname);
		fd = creat(nm,0644);
		write(fd,&rwho,len);
		close(fd);
	}
	close(sock);
	return 0;
}
